import os
import torch

from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from dataclasses import dataclass
from cftool.dist import Parallel
from cftool.misc import shallow_copy_dict
from cftool.misc import Serializer
from cftool.misc import ISerializable
from cftool.misc import DataClassBase
from cftool.misc import PureFromInfoMixin

from .task import Task
from ...data import MLData
from ...schema import DLConfig
from ...pipeline import DLEvaluationPipeline


def _task(
    task: Task,
    execute: str,
    workspace: str,
    cuda: Optional[Union[int, str]] = None,
) -> None:
    task_folder = os.path.join(workspace, "__task__")
    task.run(execute, task_folder, cuda)


def inject_distributed_tqdm_kwargs(
    i: int,
    num_jobs: int,
    config: DLConfig,
) -> None:
    tqdm_settings = config.tqdm_settings or {}
    tqdm_settings.setdefault("use_tqdm", True)
    tqdm_settings.setdefault("use_step_tqdm", False)
    tqdm_settings.setdefault("in_distributed", True)
    tqdm_settings.setdefault("tqdm_position", i % (num_jobs or 1) + 1)
    tqdm_settings.setdefault("tqdm_desc", f"epoch (task {i})")
    config.tqdm_settings = tqdm_settings


delimiter = "$^_^$"


def tuple_key_to_str_key(d: Dict[Tuple[str, str], Any]) -> Dict[str, Any]:
    return {delimiter.join(k): v for k, v in d.items()}


def str_key_to_tuple_key(d: Dict[str, Any]) -> Dict[Tuple[str, str], Any]:
    return {tuple(k.split(delimiter)): v for k, v in d.items()}  # type: ignore


@dataclass
class ExperimentResults(DataClassBase):
    workspaces: List[str]
    workspace_keys: List[Tuple[str, str]]
    pipelines: Optional[List[DLEvaluationPipeline]]


class Experiment(PureFromInfoMixin, ISerializable):
    d: Dict[str, Type["Experiment"]] = {}
    tasks_folder = "__tasks__"
    default_root_workspace = "_experiment"

    def __init__(
        self,
        *,
        num_jobs: int = 1,
        use_cuda: bool = True,
        available_cuda_list: Optional[List[int]] = None,
        resource_config: Optional[Dict[str, Any]] = None,
    ):
        use_cuda = use_cuda and torch.cuda.is_available()
        if available_cuda_list is None and not use_cuda:
            available_cuda_list = []
        self.num_jobs = num_jobs
        self.use_cuda = use_cuda
        self.cuda_list = available_cuda_list
        self.resource_config = resource_config or {}
        self.tasks: Dict[Tuple[str, str], Task] = {}
        self.key_indices: Dict[Tuple[str, str], int] = {}
        self.executes: Dict[Tuple[str, str], str] = {}
        self.workspaces: Dict[Tuple[str, str], str] = {}
        self.results: Optional[ExperimentResults] = None

    @staticmethod
    def data_folder(workspace: Optional[str] = None) -> str:
        workspace = workspace or Experiment.default_root_workspace
        return os.path.join(workspace, "__data__")

    @staticmethod
    def dump_data(
        data: MLData,
        *,
        workspace: Optional[str] = None,
        data_folder: Optional[str] = None,
    ) -> str:
        if data_folder is None:
            data_folder = Experiment.data_folder(workspace)
        os.makedirs(data_folder, exist_ok=True)
        Serializer.save(data_folder, data)
        return data_folder

    @staticmethod
    def fetch_data(
        *,
        workspace: Optional[str] = None,
        data_folder: Optional[str] = None,
    ) -> MLData:
        if data_folder is None:
            data_folder = Experiment.data_folder(workspace)
        return Serializer.load(data_folder, MLData)

    @staticmethod
    def workspace(workspace_key: Tuple[str, str], root_workspace: str) -> str:
        workspace = os.path.join(root_workspace, *workspace_key)
        return os.path.abspath(workspace)

    @property
    def num_tasks(self) -> int:
        return len(self.tasks)

    def add_task(
        self,
        data: Optional[MLData] = None,
        *,
        model: str = "fcnn",
        execute: str = "basic",
        root_workspace: Optional[str] = None,
        workspace_key: Optional[Tuple[str, str]] = None,
        config: Optional[DLConfig] = None,
        data_folder: Optional[str] = None,
        run_command: Optional[str] = None,
        task_meta_kwargs: Optional[Dict[str, Any]] = None,
    ) -> str:
        if workspace_key is None:
            counter = 0
            while True:
                workspace_key = model, str(counter)
                if workspace_key not in self.tasks:
                    break
                counter += 1
        if workspace_key in self.tasks:
            raise ValueError(f"task already exists with '{workspace_key}'")
        root_workspace = root_workspace or self.default_root_workspace
        workspace = self.workspace(workspace_key, root_workspace)
        new_idx = len(self.tasks)
        if config is not None:
            config: DLConfig = config.copy()  # type: ignore
            config.workspace = workspace
            config.create_sub_workspace = False
            config.model_name = model
            inject_distributed_tqdm_kwargs(new_idx, self.num_jobs, config)
        if data_folder is None and data is not None:
            data_folder = self.dump_data(data, workspace=workspace)
        new_task = Task.init(
            config,
            run_command,
            model=model,
            workspace=workspace,
            data_folder=None if data_folder is None else os.path.abspath(data_folder),
            **shallow_copy_dict(task_meta_kwargs or {}),
        )
        self.tasks[workspace_key] = new_task
        self.key_indices[workspace_key] = new_idx
        self.executes[workspace_key] = execute
        self.workspaces[workspace_key] = workspace
        return workspace

    def run_tasks(
        self,
        *,
        use_tqdm: bool = True,
        task_loader: Optional[Callable[[str], DLEvaluationPipeline]] = None,
        **parallel_kwargs: Any,
    ) -> ExperimentResults:
        resource_config = shallow_copy_dict(self.resource_config)
        gpu_config = resource_config.setdefault("gpu_config", {})
        gpu_config["available_cuda_list"] = self.cuda_list
        parallel = Parallel(
            self.num_jobs,
            use_tqdm=use_tqdm,
            use_cuda=self.use_cuda,
            resource_config=resource_config,
            **parallel_kwargs,
        )
        sorted_workspace_keys = sorted(self.tasks, key=self.key_indices.get)  # type: ignore
        sorted_tasks = [self.tasks[key] for key in sorted_workspace_keys]
        sorted_executes = [self.executes[key] for key in sorted_workspace_keys]
        sorted_workspaces = [self.workspaces[key] for key in sorted_workspace_keys]
        parallel(_task, sorted_tasks, sorted_executes, sorted_workspaces)
        if task_loader is None:
            pipelines = None
        else:
            pipelines = list(map(task_loader, sorted_workspaces))
        self.results = ExperimentResults(
            sorted_workspaces,
            sorted_workspace_keys,
            pipelines,
        )
        return self.results

    def to_info(self) -> Dict[str, Any]:
        tasks = {k: v.to_pack().asdict() for k, v in self.tasks.items()}
        return dict(
            num_jobs=self.num_jobs,
            use_cuda=self.use_cuda,
            cuda_list=self.cuda_list,
            resource_config=self.resource_config,
            tasks=tuple_key_to_str_key(tasks),
            key_indices=tuple_key_to_str_key(self.key_indices),
            executes=tuple_key_to_str_key(self.executes),
            workspaces=tuple_key_to_str_key(self.workspaces),
            results=None if self.results is None else self.results.asdict(),
        )

    def from_info(self, info: Dict[str, Any]) -> None:
        super().from_info(info)
        self.tasks = str_key_to_tuple_key(self.tasks)  # type: ignore
        self.tasks = {k: Task.from_pack(v) for k, v in self.tasks.items()}
        self.key_indices = str_key_to_tuple_key(self.key_indices)  # type: ignore
        self.executes = str_key_to_tuple_key(self.executes)  # type: ignore
        self.workspaces = str_key_to_tuple_key(self.workspaces)  # type: ignore
        if self.results is not None:
            workspace_keys = list(map(tuple, self.results["workspace_keys"]))
            self.results["workspace_keys"] = workspace_keys
            self.results = ExperimentResults(**self.results)


Experiment.register("experiment")(Experiment)


__all__ = [
    "inject_distributed_tqdm_kwargs",
    "Experiment",
    "ExperimentResults",
]



from .schema import *
from .attn import *
from .fnet import *
from .mixer import *
from .vqgan import *
from .vanilla import *
from .backbone import *
from .perceiver import *
from .pool_former import *
from .transformer import *



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional

from .base import MLModel
from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings
from ...modules.blocks import Linear
from ...modules.blocks import MixedStackedEncoder


class MixedStackedModel(MLModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        *,
        latent_dim: int,
        token_mixing_type: str,
        token_mixing_config: Optional[Dict[str, Any]] = None,
        channel_mixing_type: str = "ff",
        channel_mixing_config: Optional[Dict[str, Any]] = None,
        num_layers: int = 4,
        dropout: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
        feedforward_dim_ratio: float = 1.0,
        use_head_token: bool = False,
        head_pooler: Optional[str] = "mean",
        use_positional_encoding: bool = False,
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )
        if self.encoder is not None:
            input_dim += self.encoder.dim_increment
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_history = num_history
        self.to_mixed_stacked = Linear(input_dim, latent_dim)
        self.mixed_stacked = MixedStackedEncoder(
            latent_dim,
            num_history,
            token_mixing_type=token_mixing_type,
            token_mixing_config=token_mixing_config,
            channel_mixing_type=channel_mixing_type,
            channel_mixing_config=channel_mixing_config,
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
            latent_dim_ratio=feedforward_dim_ratio,
            use_head_token=use_head_token,
            head_pooler=head_pooler,
            use_positional_encoding=use_positional_encoding,
        )
        self.head = Linear(latent_dim, output_dim)

    def forward(self, net: Tensor) -> Tensor:
        net = self.to_mixed_stacked(net)
        net = self.mixed_stacked(net)
        net = self.head(net)
        return net


@MLModel.register("fnet")
class FNet(MixedStackedModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        *,
        latent_dim: int = 256,
        num_layers: int = 4,
        dropout: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            input_dim,
            output_dim,
            num_history,
            latent_dim=latent_dim,
            token_mixing_type="fourier",
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
            feedforward_dim_ratio=1.0,
            use_head_token=False,
            use_positional_encoding=False,
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )


@MLModel.register("mixer")
class Mixer(MixedStackedModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        *,
        latent_dim: int = 256,
        num_layers: int = 4,
        dropout: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            input_dim,
            output_dim,
            num_history,
            latent_dim=latent_dim,
            token_mixing_type="mlp",
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
            feedforward_dim_ratio=1.0,
            use_head_token=False,
            use_positional_encoding=False,
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )


@MLModel.register("transformer")
class Transformer(MixedStackedModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        *,
        latent_dim: int = 256,
        num_layers: int = 4,
        dropout: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
        feedforward_dim_ratio: float = 4.0,
        use_head_token: bool = False,
        attention_config: Optional[Dict[str, Any]],
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        attention_config = attention_config or {}
        attention_config.setdefault("bias", False)
        attention_config.setdefault("num_heads", 8)
        super().__init__(
            input_dim,
            output_dim,
            num_history,
            latent_dim=latent_dim,
            token_mixing_type="attention",
            token_mixing_config=attention_config,
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
            feedforward_dim_ratio=feedforward_dim_ratio,
            use_head_token=use_head_token,
            use_positional_encoding=True,
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )


@MLModel.register("pool_former")
class PoolFormer(MixedStackedModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        *,
        latent_dim: int = 256,
        num_layers: int = 4,
        dropout: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            input_dim,
            output_dim,
            num_history,
            latent_dim=latent_dim,
            token_mixing_type="pool",
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
            feedforward_dim_ratio=1.0,
            use_head_token=False,
            use_positional_encoding=False,
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )


__all__ = [
    "FNet",
    "Mixer",
    "Transformer",
    "PoolFormer",
]



import torch

from torch import nn
from typing import Any
from typing import Callable
from typing import Iterator
from cftool.types import tensor_dict_type


class DropNoGradStatesMixin:
    def state_dict(
        self,
        *,
        destination: Any = None,
        prefix: str = "",
        keep_vars: bool = False,
    ) -> tensor_dict_type:
        states = super().state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)  # type: ignore
        for key, _ in self.named_buffers():  # type: ignore
            if states.pop(key, None) is None:
                states.pop(f"core.{key}")
        for key, value in self.named_parameters():  # type: ignore
            if not value.requires_grad:
                if states.pop(key, None) is None:
                    states.pop(f"core.{key}")
        return states

    def load_state_dict(
        self,
        state_dict: tensor_dict_type,
        strict: bool = True,
    ) -> None:
        with torch.no_grad():
            for key, value in self.named_parameters():  # type: ignore
                if value.requires_grad:
                    loaded_value = state_dict.get(key)
                    if strict and loaded_value is None:
                        raise ValueError(f"value for '{key}' is missing")
                    value.data.copy_(loaded_value)



import torch

import torch.nn as nn

from abc import abstractmethod
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from typing import Tuple
from typing import Optional
from torch.nn import Module
from torch.nn import ModuleList
from cftool.misc import WithRegister
from cftool.array import squeeze

from .convs import get_conv_blocks
from .convs import Conv2d
from .norms import NormFactory
from .customs import Linear
from .attentions import Attention
from ...misc.toolkit import get_device
from ...misc.toolkit import eval_context


class PreNorm(Module):
    def __init__(
        self,
        *dims: int,
        module: Module,
        norm_type: Optional[str] = "layer",
        norm_kwargs: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        if norm_kwargs is None:
            norm_kwargs = {}
        self.norms = ModuleList([])
        for dim in dims:
            self.norms.append(NormFactory(norm_type).make(dim, **norm_kwargs))
        self.module = module

    def forward(self, *xs: Tensor, **kwargs: Any) -> Tensor:
        x_list = [norm(x) for x, norm in zip(xs, self.norms)]
        if not issubclass(self.module.__class__, Attention):
            return self.module(*x_list, **kwargs)
        if len(x_list) == 1:
            x_list = [x_list[0]] * 3
        elif len(x_list) == 2:
            x_list.append(x_list[1])
        if len(x_list) != 3:
            raise ValueError("there should be three inputs for `Attention`")
        return_attention = kwargs.pop("return_attention", False)
        attention_outputs = self.module(*x_list, **kwargs)
        if return_attention:
            return attention_outputs.weights
        return attention_outputs.output


class ChannelPadding(Module):
    def __init__(
        self,
        in_channels: int,
        latent_channels: int,
        map_dim: Optional[int] = None,
        *,
        is_1d: bool = False,
        num_classes: Optional[int] = None,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.latent_channels = latent_channels
        self.map_dim = map_dim
        self.is_global = map_dim is None
        self.is_conditional = num_classes is not None
        if self.is_global:
            map_dim = 1
        token_shape = (num_classes or 1), latent_channels, map_dim, map_dim
        self.channel_padding = nn.Parameter(torch.randn(*token_shape))  # type: ignore
        in_nc = in_channels + latent_channels
        if is_1d:
            self.mapping = Linear(in_nc, in_channels, bias=False)
        else:
            self.mapping = Conv2d(in_nc, in_channels, kernel_size=1, bias=False)

    def forward(self, net: Tensor, labels: Optional[Tensor] = None) -> Tensor:
        if not self.is_conditional:
            padding = self.channel_padding.repeat(net.shape[0], 1, 1, 1)
        else:
            if labels is None:
                msg = "`labels` should be provided in conditional `ChannelPadding`"
                raise ValueError(msg)
            padding = self.channel_padding[labels.view(-1)]
        if self.is_global:
            if len(net.shape) == 2:
                padding = squeeze(padding)
            else:
                padding = padding.repeat(1, 1, *net.shape[-2:])
        net = torch.cat([net, padding], dim=1)
        net = self.mapping(net)
        return net

    def extra_repr(self) -> str:
        dim_str = f"{self.in_channels}+{self.latent_channels}"
        map_dim_str = "global" if self.is_global else f"{self.map_dim}x{self.map_dim}"
        return f"{dim_str}, {map_dim_str}"


to_patches: Dict[str, Type["ImgToPatches"]] = {}


class ImgToPatches(Module, WithRegister["ImgToPatches"]):
    d = to_patches

    def __init__(
        self,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_dim: int,
    ):
        super().__init__()
        self.img_size = img_size
        self.patch_size = patch_size
        self.in_channels = in_channels
        self.latent_dim = latent_dim

    @abstractmethod
    def forward(self, net: Tensor, *, determinate: bool = False) -> Tuple[Tensor, Any]:
        """should return patches and its hw"""

    @property
    def num_patches(self) -> int:
        shape = 1, self.in_channels, self.img_size, self.img_size
        with eval_context(self):
            net = self.forward(torch.zeros(*shape, device=get_device(self)))[0]
        return net.shape[1]

    @staticmethod
    def _flatten(net: Tensor, determinate: bool) -> Tuple[Tensor, Any]:
        c, h, w = net.shape[1:]
        if determinate:
            c, h, w = map(int, [c, h, w])
        net = net.view(-1, c, h * w).transpose(1, 2).contiguous()
        return net, (h, w)


@ImgToPatches.register("vanilla")
class VanillaPatchEmbed(ImgToPatches):
    def __init__(
        self,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_dim: int = 128,
        **conv_kwargs: Any,
    ):
        super().__init__(img_size, patch_size, in_channels, latent_dim)
        if img_size % patch_size != 0:
            raise ValueError(
                f"`img_size` ({img_size}) should be "
                f"divisible by `patch_size` ({patch_size})"
            )
        self.img_size = img_size
        self.patch_size = patch_size
        self.in_channels = in_channels
        self.latent_dim = latent_dim
        self.projection = Conv2d(
            in_channels,
            latent_dim,
            kernel_size=patch_size,
            stride=patch_size,
            padding=0,
            **conv_kwargs,
        )

    def forward(self, net: Tensor, *, determinate: bool = False) -> Tuple[Tensor, Any]:
        net = self.projection(net)
        return self._flatten(net, determinate)


@ImgToPatches.register("overlap")
class OverlapPatchEmbed(ImgToPatches):
    def __init__(
        self,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_dim: int = 768,
        stride: int = 4,
        **conv_kwargs: Any,
    ):
        super().__init__(img_size, patch_size, in_channels, latent_dim)
        self.conv = Conv2d(
            in_channels,
            latent_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=(patch_size // 2, patch_size // 2),
            **conv_kwargs,
        )
        self.norm = nn.LayerNorm(latent_dim)

    def forward(self, net: Tensor, *, determinate: bool = False) -> Tuple[Tensor, Any]:
        net = self.conv(net)
        net, hw = self._flatten(net, determinate)
        net = self.norm(net)
        return net, hw


@ImgToPatches.register("conv")
class ConvPatchEmbed(ImgToPatches):
    def __init__(
        self,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_channels: int = 64,
        latent_dim: int = 384,
        padding: Optional[int] = None,
        stride: Optional[int] = None,
        bias: bool = False,
        num_layers: int = 2,
        activation: str = "relu",
    ):
        super().__init__(img_size, patch_size, in_channels, latent_dim)
        latent_channels_list = [latent_channels] * (num_layers - 1)
        num_channels_list = [in_channels] + latent_channels_list + [latent_dim]
        if padding is None:
            padding = max(1, patch_size // 2)
        if stride is None:
            stride = max(1, (patch_size // 2) - 1)
        self.conv = nn.Sequential(
            *[
                nn.Sequential(
                    *get_conv_blocks(
                        num_channels_list[i],
                        num_channels_list[i + 1],
                        patch_size,
                        stride,
                        bias=bias,
                        activation=activation,
                        padding=padding,
                    ),
                    nn.MaxPool2d(3, 2, 1),
                )
                for i in range(num_layers)
            ]
        )

    def forward(self, net: Tensor, *, determinate: bool = False) -> Tuple[Tensor, Any]:
        net = self.conv(net)
        return self._flatten(net, determinate)


__all__ = [
    "PreNorm",
    "ChannelPadding",
    "ImgToPatches",
    "VanillaPatchEmbed",
]



import torch
import random

import torch.nn as nn

from abc import abstractmethod
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Callable
from typing import Optional
from cftool.types import tensor_dict_type

from ...schemas import CustomTrainStep
from ...schemas import CustomTrainStepLoss
from ...schemas import ModelWithCustomSteps
from ....schema import TrainerState
from ....schema import MetricsOutputs
from ....constants import LOSS_KEY
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import PREDICTIONS_KEY
from ...schemas.cv import GaussianGeneratorMixin
from ....misc.toolkit import get_device
from ....misc.toolkit import mode_context
from ....losses.gan import GANLoss
from ....losses.gan import GANTarget
from ....losses.gan import DiscriminatorOutput


class GeneratorStep(CustomTrainStep):
    def loss_fn(
        self,
        m: "IOneStageGAN",
        state: Optional[TrainerState],
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        g_losses = m.g_losses(batch, forward_results)
        g_loss = g_losses.pop(LOSS_KEY)
        g_loss_dict = {k: v.item() for k, v in g_losses.items()}
        g_loss_dict["g"] = g_loss.item()
        return CustomTrainStepLoss(g_loss, g_loss_dict)


class DiscriminatorStep(CustomTrainStep):
    def loss_fn(
        self,
        m: "IOneStageGAN",
        state: Optional[TrainerState],
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        with torch.no_grad():
            detached_forward = {k: v.detach() for k, v in forward_results.items()}
        d_losses = m.d_losses(batch, detached_forward)
        d_loss = d_losses.pop(LOSS_KEY)
        d_loss_dict = {k: v.item() for k, v in d_losses.items()}
        d_loss_dict["g"] = d_loss.item()
        return CustomTrainStepLoss(d_loss, d_loss_dict)


class IOneStageGAN(ModelWithCustomSteps, metaclass=ABCMeta):
    # abstract

    @property
    @abstractmethod
    def g_parameters(self) -> List[nn.Parameter]:
        pass

    @property
    @abstractmethod
    def d_parameters(self) -> List[nn.Parameter]:
        pass

    @abstractmethod
    def g_losses(
        self,
        batch: tensor_dict_type,
        forward: tensor_dict_type,
    ) -> tensor_dict_type:
        pass

    @abstractmethod
    def d_losses(
        self,
        batch: tensor_dict_type,
        detached_forward: tensor_dict_type,
    ) -> tensor_dict_type:
        pass

    # inheritance

    @property
    def train_steps(self) -> List[CustomTrainStep]:
        return [
            GeneratorStep("g_parameters"),
            DiscriminatorStep("d_parameters"),
        ]

    def evaluate(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"],
        weighted_loss_score_fn: Callable[[Dict[str, float]], float],
        forward_kwargs: Dict[str, Any],
    ) -> MetricsOutputs:
        forward = self.run(batch_idx, batch)
        g_losses = self.g_losses(batch, forward)
        # in evaluate step, all tensors are already detached
        d_losses = self.d_losses(batch, forward)
        g_loss = g_losses.pop(LOSS_KEY).item()
        d_loss = d_losses.pop(LOSS_KEY).item()
        loss = g_loss + d_loss
        loss_dict = {"g": g_loss, "d": d_loss, LOSS_KEY: loss}
        loss_dict.update({k: v.item() for k, v in g_losses.items()})
        loss_dict.update({k: v.item() for k, v in d_losses.items()})
        score = weighted_loss_score_fn(loss_dict)
        return MetricsOutputs(score, loss_dict, {k: False for k in loss_dict})

    def summary_forward(self, batch: tensor_dict_type) -> None:
        self.g_losses(batch, self.run(0, batch))

    # api

    def build_loss(
        self,
        *,
        num_classes: Optional[int] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.num_classes = num_classes
        self.gan_mode = gan_mode
        self.gan_loss = GANLoss(gan_mode)
        if gan_loss_config is None:
            gan_loss_config = {}
        self.lambda_gp = gan_loss_config.get("lambda_gp", 10.0)


class IGenerator(nn.Module, metaclass=ABCMeta):
    @abstractmethod
    def decode(self, batch: tensor_dict_type) -> Tensor:
        pass


class IDiscriminator(nn.Module):
    @abstractmethod
    def forward(self, net: Tensor) -> DiscriminatorOutput:
        pass


class IVanillaGAN(IOneStageGAN, GaussianGeneratorMixin, metaclass=ABCMeta):
    generator: IGenerator
    discriminator: IDiscriminator

    # inheritance

    @property
    def can_reconstruct(self) -> bool:
        return False

    @property
    def g_parameters(self) -> List[nn.Parameter]:
        return list(self.generator.parameters())

    @property
    def d_parameters(self) -> List[nn.Parameter]:
        return list(self.discriminator.parameters())

    def get_forward_args(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        return batch[INPUT_KEY], batch.get(LABEL_KEY)

    def forward(self, net: Tensor, labels: Optional[Tensor]) -> Tensor:
        z = torch.randn(len(net), self.latent_dim, device=get_device(self))
        return self.decode(z, labels=labels)

    def decode(self, z: Tensor, *, labels: Optional[Tensor], **kwargs: Any) -> Tensor:
        batch = {INPUT_KEY: z, LABEL_KEY: labels}
        net = self.generator.decode(batch, **kwargs)
        return net

    def g_losses(
        self,
        batch: tensor_dict_type,
        forward: tensor_dict_type,
    ) -> tensor_dict_type:
        labels = batch.get(LABEL_KEY)
        if labels is not None:
            labels = labels.view(-1)
        sampled = forward[PREDICTIONS_KEY]
        pred_fake = self.discriminator(sampled)
        loss_g = self.gan_loss(pred_fake, GANTarget(True, labels))
        return {LOSS_KEY: loss_g}

    def d_losses(
        self,
        batch: tensor_dict_type,
        detached_forward: tensor_dict_type,
    ) -> tensor_dict_type:
        net = batch[INPUT_KEY]
        labels = batch.get(LABEL_KEY)
        if labels is not None:
            labels = labels.view(-1)
        sampled_tensor = detached_forward[PREDICTIONS_KEY]
        pred_real = self.discriminator(net)
        loss_d_real = self.gan_loss(pred_real, GANTarget(True, labels))
        pred_fake = self.discriminator(sampled_tensor)
        loss_d_fake = self.gan_loss(pred_fake, GANTarget(False, labels))
        d_loss = 0.5 * (loss_d_fake + loss_d_real)
        losses = {"d_fake": loss_d_fake, "d_real": loss_d_real}
        if self.gan_mode == "wgangp":
            eps = random.random()
            merged = eps * net + (1.0 - eps) * sampled_tensor
            with mode_context(self.discriminator, to_train=None, use_grad=True):
                pred_merged = self.discriminator(merged.requires_grad_(True)).output  # type: ignore
                loss_gp = self.gan_loss.loss(merged, pred_merged)
            d_loss = d_loss + self.lambda_gp * loss_gp
            losses["d_gp"] = loss_gp
        losses[LOSS_KEY] = d_loss
        return losses


__all__ = [
    "IOneStageGAN",
    "IVanillaGAN",
]



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from typing import Callable
from typing import Optional
from cftool.misc import check_requires
from cftool.misc import shallow_copy_dict
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type

from ....data import ArrayDictData
from ....types import texts_type
from ....schema import IDLModel
from ....schema import IInference
from ....schema import TrainerState

try:
    from transformers import AutoModel
    from transformers import AutoTokenizer
except:
    AutoModel = AutoTokenizer = None


@IDLModel.register("hugging_face")
class HuggingFaceModel(IDLModel):
    tokenizer_base: Type = AutoTokenizer
    model_base: Type = AutoModel
    forward_fn_name: str = "forward"

    def __init__(self, model: str):
        if AutoModel is None:
            raise ValueError("`transformers` is needed for `HuggingFaceModel`")
        super().__init__()
        self.tokenizer = self.tokenizer_base.from_pretrained(model)
        self.model = self.model_base.from_pretrained(model)

    def model_forward(self, batch: tensor_dict_type) -> Any:
        fn = getattr(self.model, self.forward_fn_name)
        batch = shallow_copy_dict(batch)
        pop_keys = [k for k in batch.keys() if not check_requires(fn, k, False)]
        for k in pop_keys:
            batch.pop(k)
        return fn(**batch)

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        return self.model_forward(batch)

    def inference(
        self,
        texts: texts_type,
        use_tqdm: bool = True,
        **kwargs: Any,
    ) -> np_dict_type:
        x = self.tokenizer(texts, padding=True, return_tensors="pt")
        loader = ArrayDictData.init().fit(x).get_loaders()[0]
        inference = IInference(model=self)
        outputs = inference.get_outputs(loader, use_tqdm=use_tqdm, **kwargs)
        return outputs.forward_results

    def to_onnx(  # type: ignore
        self,
        export_folder: str,
        sample_text: str,
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        max_length: Optional[int] = None,
        onnx_file: str = "model.onnx",
        opset: int = 11,
        simplify: bool = True,
        forward_fn: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
        output_names: Optional[List[str]] = None,
        num_samples: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> "IDLModel":
        if max_length is None:
            input_sample = self.tokenizer(sample_text, return_tensors="pt")
        else:
            input_sample = self.tokenizer(
                sample_text,
                padding="max_length",
                truncation=True,
                max_length=max_length,
                return_tensors="pt",
            )
        return super().to_onnx(
            export_folder,
            input_sample,
            dynamic_axes,
            onnx_file=onnx_file,
            opset=opset,
            simplify=simplify,
            forward_fn=forward_fn,
            output_names=output_names,
            num_samples=num_samples,
            verbose=verbose,
            **kwargs,
        )


__all__ = [
    "HuggingFaceModel",
]



import torch

from typing import List
from typing import NamedTuple

try:
    from transformers import GPT2Tokenizer
    from transformers import GPT2LMHeadModel
except:
    GPT2Tokenizer = GPT2LMHeadModel = None


class PromptEnhanceConfig(NamedTuple):
    temperature: float = 0.9
    top_k: int = 8
    max_length: int = 76
    repitition_penalty: float = 1.2
    num_return_sequences: int = 1
    comma_mode: bool = False


class PromptEnhanceAPI:
    def __init__(self, device: str = "cpu", *, use_half: bool = False) -> None:
        if GPT2Tokenizer is None or GPT2LMHeadModel is None:
            raise ValueError("`trainsformers` is required for `PromptEnhanceAPI`")
        self.tokenizer = GPT2Tokenizer.from_pretrained("distilgpt2")
        self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})
        version = "FredZhang7/distilgpt2-stable-diffusion-v2"
        model = GPT2LMHeadModel.from_pretrained(version)
        self.model = model.half().to(device)
        self.device = device

    def to(self, device: str, *, use_half: bool = False) -> None:
        self.device = device
        self.model.to(device)

    @torch.no_grad()
    def enhance(self, prompt: str, config: PromptEnhanceConfig) -> List[str]:
        input_ids = self.tokenizer(prompt, return_tensors="pt").input_ids
        input_ids = input_ids.to(self.device)
        kw = dict(
            do_sample=True,
            temperature=config.temperature,
            top_k=config.top_k,
            max_length=config.max_length,
            num_return_sequences=config.num_return_sequences,
            repetition_penalty=config.repitition_penalty,
            early_stopping=True,
        )
        if not config.comma_mode:
            kw.update(dict(penalty_alpha=0.6, no_repeat_ngram_size=1))
        outputs = self.model.generate(input_ids, **kw)
        return [self.tokenizer.decode(out, skip_special_tokens=True) for out in outputs]


__all__ = [
    "PromptEnhanceConfig",
    "PromptEnhanceAPI",
]



import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Tuple

from .schema import EncoderMixin
from ....modules.blocks import make_attention
from ....modules.blocks import HijackConv2d
from ....modules.blocks import ResidualBlockWithTimeEmbedding


class Downsample(nn.Module):
    def __init__(self, in_channels: int, with_conv: bool):
        super().__init__()
        if not with_conv:
            self.conv = None
        else:
            self.conv = HijackConv2d(in_channels, in_channels, 3, 2, 0)

    def forward(self, net: Tensor) -> Tensor:
        if self.conv is None:
            return F.avg_pool2d(net, kernel_size=2, stride=2)
        pad = (0, 1, 0, 1)
        net = F.pad(net, pad, mode="constant", value=0)
        net = self.conv(net)
        return net


@EncoderMixin.register("attention")
class AttentionEncoder(nn.Module, EncoderMixin):
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        inner_channels: int,
        latent_channels: int,
        channel_multipliers: Tuple[int, ...],
        *,
        num_res_blocks: int,
        attention_resolutions: Tuple[int, ...] = (),
        dropout: float = 0.0,
        downsample_with_conv: bool = True,
        attention_type: str = "spatial",
    ):
        super().__init__()
        self.img_size = img_size
        self.in_channels = in_channels
        self.inner_channels = inner_channels
        self.latent_channels = latent_channels
        self.num_downsample = len(channel_multipliers)
        self.num_res_blocks = num_res_blocks
        make_res_block = lambda in_c, out_c: ResidualBlockWithTimeEmbedding(
            in_c,
            out_c,
            dropout=dropout,
            time_embedding_channels=0,
        )
        # in conv
        blocks = [HijackConv2d(in_channels, inner_channels, 3, 1, 1)]
        # downsample
        current_resolution = img_size
        in_channel_multipliers = (1,) + tuple(channel_multipliers)
        self.in_channel_multipliers = in_channel_multipliers
        in_nc = 0
        for i in range(self.num_downsample):
            in_nc = inner_channels * in_channel_multipliers[i]
            out_nc = inner_channels * channel_multipliers[i]
            for _ in range(self.num_res_blocks):
                blocks.append(make_res_block(in_nc, out_nc))
                in_nc = out_nc
                if current_resolution in attention_resolutions:
                    blocks.append(make_attention(in_nc, attention_type))
            if i != self.num_downsample - 1:
                blocks.append(Downsample(in_nc, downsample_with_conv))
                current_resolution //= 2
        # residual
        blocks += [
            make_res_block(in_nc, in_nc),
            make_attention(in_nc, attention_type),
            make_res_block(in_nc, in_nc),
        ]
        # output
        blocks += [
            nn.GroupNorm(num_groups=32, num_channels=in_nc, eps=1.0e-6, affine=True),
            nn.SiLU(),
            HijackConv2d(in_nc, latent_channels, 3, 1, 1),
        ]
        # construct
        self.encoder = nn.Sequential(*blocks)

    def forward(self, net: Tensor) -> Tensor:
        return self.encoder(net)


__all__ = [
    "AttentionEncoder",
]



import math
import torch

import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Tuple
from typing import Optional
from torch.fft import fft
from cftool.misc import safe_execute

from .schema import TokenMixerBase
from .channel_mixers import FeedForward
from ..common import Lambda
from ..attentions import Attention


@TokenMixerBase.register("mlp")
class MLPTokenMixer(TokenMixerBase):
    def __init__(self, in_dim: int, num_tokens: int, *, dropout: float = 0.1):
        super().__init__(in_dim, num_tokens)
        self.net = nn.Sequential(
            Lambda(lambda x: x.transpose(1, 2), name="to_token_mixing"),
            FeedForward(num_tokens, num_tokens, dropout),
            Lambda(lambda x: x.transpose(1, 2), name="to_channel_mixing"),
        )

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
    ) -> Tensor:
        return self.net(net)


@TokenMixerBase.register("fourier")
class FourierTokenMixer(TokenMixerBase):
    def __init__(self, in_dim: int, num_tokens: int):
        super().__init__(in_dim, num_tokens)
        self.net = Lambda(lambda x: fft(fft(x, dim=-1), dim=-2).real, name="fourier")

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
    ) -> Tensor:
        return self.net(net)


@TokenMixerBase.register("attention")
class AttentionTokenMixer(TokenMixerBase):
    def __init__(
        self,
        in_dim: int,
        num_tokens: int,
        *,
        attention_type: str = "basic",
        **attention_kwargs: Any,
    ):
        super().__init__(in_dim, num_tokens)
        attention_kwargs.setdefault("bias", False)
        attention_kwargs.setdefault("num_heads", 8)
        attention_kwargs["input_dim"] = in_dim
        attention_kwargs.setdefault("is_self_attention", True)
        self.net = safe_execute(Attention.get(attention_type), attention_kwargs)

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        mask: Optional[Tensor] = None,
    ) -> Tensor:
        kw = dict(hw=hw, mask=mask, determinate=determinate)
        return self.net(net, net, net, **kw).output


@TokenMixerBase.register("pool")
class PoolTokenMixer(TokenMixerBase):
    def __init__(self, in_dim: int, num_tokens: int, *, pool_size: int = 3):
        super().__init__(in_dim, num_tokens)
        self.pool = nn.AvgPool2d(
            pool_size,
            stride=1,
            padding=pool_size // 2,
            count_include_pad=False,
        )

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
    ) -> Tensor:
        return self.pool(net) - net


@TokenMixerBase.register("rwkv")
class RWKVTokenMixer(TokenMixerBase):
    def __init__(
        self,
        layer_idx: int,
        num_layers: int,
        num_tokens: int,
        in_dim: int,
        latent_dim: int,
    ):
        super().__init__(in_dim, num_tokens)
        latent_dim = latent_dim or in_dim
        self.ci = latent_dim

        self.time_shift = nn.ZeroPad2d((0, 0, 1, -1))
        self.key = nn.Linear(in_dim, latent_dim, bias=False)
        self.value = nn.Linear(in_dim, latent_dim, bias=False)
        self.receptance = nn.Linear(in_dim, latent_dim, bias=False)
        self.output = nn.Linear(latent_dim, in_dim, bias=False)

        with torch.no_grad():
            for k, v in self.state_dict().items():
                for zero_k in ["key", "receptance", "output"]:
                    if zero_k in k:
                        nn.init.zeros_(v)
                        break
                else:
                    nn.init.orthogonal_(v)

            _0_to_1 = layer_idx / (num_layers - 1)
            _1_to_almost0 = 1.0 - (layer_idx / num_layers)

            f1_begin = 3.0
            f1_end = 1.2
            f2_begin = 0.65
            f2_end = 0.4
            decay = torch.ones(latent_dim, 1)
            first_sa_layer_id = int(num_layers > 2)
            for h in range(latent_dim):
                mul = layer_idx - first_sa_layer_id
                divider = num_layers - 1 - first_sa_layer_id
                f1 = f1_begin + mul / divider * (f1_end - f1_begin)
                f2 = f2_begin + mul / divider * (f2_end - f2_begin)
                if layer_idx == first_sa_layer_id:
                    f1 += 0.5
                elif layer_idx == num_layers - 2:
                    f2 = 0.4
                elif layer_idx == num_layers - 1:
                    f2 = 0.37
                decay[h][0] = math.pow(f2, h / (latent_dim - 1) * 7.0) * f1
            # Cl, 1
            self.info_decay = nn.Parameter(decay.log())
            # 1, T
            time_curve = torch.arange(-num_tokens + 1, 1)[None]
            self.register_buffer("time_curve", time_curve)

            x = torch.ones(1, 1, in_dim)
            for i in range(in_dim):
                x[0, 0, i] = i / in_dim
            self.time_mix_k = nn.Parameter(torch.pow(x, _1_to_almost0))
            self.time_mix_v = nn.Parameter(torch.pow(x, _1_to_almost0) + 0.3 * _0_to_1)
            self.time_mix_r = nn.Parameter(torch.pow(x, 0.5 * _1_to_almost0))

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
    ) -> Tensor:
        # B, T, C
        t = net.shape[1]
        shifted = self.time_shift(net)
        net_k = net * self.time_mix_k + shifted * (1.0 - self.time_mix_k)
        net_v = net * self.time_mix_v + shifted * (1.0 - self.time_mix_v)
        net_r = net * self.time_mix_r + shifted * (1.0 - self.time_mix_r)

        # B, Cl, T
        net_k = self.key(net_k).transpose(1, 2)
        net_v = self.value(net_v).transpose(1, 2)

        # Cl, T
        exp_w = (self.info_decay.exp() * self.time_curve).exp()
        # Cl, 1, T
        exp_w = exp_w[:, None]
        # B, Cl, T
        exp_k = net_k.exp()
        net_kv = exp_k * net_v
        # B, Cl, T
        wk = F.conv1d(F.pad(exp_k, (t - 1, 0, 0, 0)), exp_w, groups=self.ci)
        net_kv = net_kv / (wk + 1.0e-6)
        wkv = F.conv1d(F.pad(net_kv, (t - 1, 0, 0, 0)), exp_w, groups=self.ci)
        # B, T, Cl
        wkv = wkv.transpose(1, 2)
        net_r = self.receptance(net_r)
        net_r = torch.sigmoid(net_r)
        rwkv = net_r * wkv
        # B, T, C
        net = self.output(rwkv)
        return net.contiguous()


__all__ = [
    "MLPTokenMixer",
    "FourierTokenMixer",
    "AttentionTokenMixer",
    "PoolTokenMixer",
    "RWKVTokenMixer",
]



from typing import Union
from typing import Optional
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict

from .basic import SetDefaultsBlock
from .basic import SetTrainerDefaultsBlock
from ..core import Block
from ...schema import DLConfig
from ...schema import MLConfig
from ...schema import MLEncoderSettings
from ...data.ml import MLData
from ...data.blocks import ColumnTypes
from ...data.blocks import RecognizerBlock


# static blocks


@Block.register("set_ml_defaults")
class SetMLDefaultsBlock(SetDefaultsBlock):
    data: MLData

    def build(self, config: Union[DLConfig, MLConfig]) -> None:
        super().build(config)
        if not isinstance(config, MLConfig):
            return
        self._infer_encoder_settings(config)
        b_recognizer = self.b_recognizer
        if b_recognizer is not None:
            config.index_mapping = shallow_copy_dict(b_recognizer.index_mapping)

    @property
    def b_recognizer(self) -> Optional[RecognizerBlock]:
        return self.data.processor.try_get_block(RecognizerBlock)

    def _infer_encoder_settings(self, config: MLConfig) -> None:
        if config.encoder_settings is not None:
            return
        if not config.infer_encoder_settings:
            return
        if self.data is None:
            print_warning(
                "`infer_encoder_settings` is set to True "
                "but `data` is not provided, it will take no effect."
            )
            return
        b_recognizer = self.b_recognizer
        if b_recognizer is None:
            print_warning(
                "`infer_encoder_settings` is set to True "
                "but `RecognizerBlock` is not provided, it will take no effect"
            )
            return
        encoder_settings = {}
        for original_idx in b_recognizer.index_mapping:
            if b_recognizer.feature_types[original_idx] == ColumnTypes.CATEGORICAL:
                dim = b_recognizer.num_unique_features[original_idx]
                encoder_settings[original_idx] = MLEncoderSettings(dim)
        config.encoder_settings = encoder_settings
        d = {k: v.asdict() for k, v in encoder_settings.items()}
        self._defaults["encoder_settings"] = d


@Block.register("set_ml_trainer_defaults")
class SetMLTrainerDefaultsBlock(SetTrainerDefaultsBlock):
    def build(self, config: DLConfig) -> None:
        if config.monitor_names is None:
            config.monitor_names = ["mean_std", "plateau"]
            self._defaults["monitor_names"] = ["mean_std", "plateau"]
        super().build(config)


__all__ = [
    "SetMLDefaultsBlock",
    "SetMLTrainerDefaultsBlock",
]



import torch

import numpy as np

from abc import abstractmethod
from abc import ABCMeta

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional
from typing import Protocol

from .utils import append_dims
from .utils import append_zero
from .schema import ISampler
from .schema import IQSampler
from .schema import IDiffusion
from .schema import UncondSamplerMixin
from ..utils import cond_type
from ..utils import extract_to

try:
    from scipy import integrate
except:
    integrate = None


def to_d(image: Tensor, sigma: Tensor, denoised: Tensor) -> Tensor:
    return (image - denoised) / append_dims(sigma, image.ndim)


def get_ancestral_step(
    sigma_from: float,
    sigma_to: float,
    eta: float = 1.0,
) -> Tuple[float, float]:
    if not eta:
        return sigma_to, 0.0
    sigma_up = min(
        sigma_to,
        eta
        * (sigma_to**2 * (sigma_from**2 - sigma_to**2) / sigma_from**2) ** 0.5,
    )
    sigma_down = (sigma_to**2 - sigma_up**2) ** 0.5
    return sigma_down, sigma_up


class IGetDenoised(Protocol):
    def __call__(self, img: Tensor, sigma: Tensor) -> Tensor:
        pass


class KQSampler(IQSampler):
    def q_sample(
        self,
        net: Tensor,
        timesteps: Tensor,
        noise: Optional[Tensor] = None,
    ) -> Tensor:
        w_noise = extract_to(self.sigmas, timesteps, net.ndim)
        if noise is None:
            noise = torch.randn_like(net)
        net = net + w_noise * noise
        return net

    def reset_buffers(self, sigmas: Tensor) -> None:  # type: ignore
        self.sigmas = sigmas


class KSamplerMixin(ISampler, UncondSamplerMixin, metaclass=ABCMeta):
    def __init__(
        self,
        model: IDiffusion,
        *,
        default_quantize: bool = False,
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        default_steps: int = 25,
    ):
        if model.parameterization not in ("eps", "v"):
            msg = "only `v` / `eps` parameterization is supported in `k_samplers`"
            raise ValueError(msg)
        super().__init__(model)
        self.default_quantize = default_quantize
        self.unconditional_cond = unconditional_cond
        self.unconditional_guidance_scale = unconditional_guidance_scale
        self.default_steps = default_steps

    # abstract

    @abstractmethod
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        get_denoised: IGetDenoised,
        *,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        **kwargs: Any,
    ) -> Tensor:
        pass

    # inheritance

    @property
    def q_sampler(self) -> KQSampler:
        return self._q_sampler

    @property
    def sample_kwargs(self) -> Dict[str, Any]:
        return dict(
            quantize=self.default_quantize,
            unconditional_cond=self.unconditional_cond,
            unconditional_guidance_scale=self.unconditional_guidance_scale,
        )

    def sample_step(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        *,
        quantize: bool = False,
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        **kwargs: Any,
    ) -> Tensor:
        if step == 0 and not self.initialized:
            self._reset_buffers(
                total_step,
                quantize,
                unconditional_cond,
                unconditional_guidance_scale,
            )
            image = image * self.sigmas[0]
        if step == total_step - 1:
            return image

        def get_denoised(img: Tensor, sigma: Tensor) -> Tensor:
            ndim = img.ndim
            sigma = sigma * s_in
            ts = self._sigma_to_t(sigma, quantize)
            if self.model.parameterization == "eps":
                c_in = append_dims(
                    1.0 / (sigma**2 + self.sigma_data**2) ** 0.5,
                    ndim,
                )
                c_out = append_dims(sigma, ndim)
                eps = self._uncond_denoise(img * c_in, ts, cond, step, total_step)
                return img - eps * c_out
            if self.model.parameterization == "v":
                c_in = append_dims(
                    1.0 / (sigma**2 + self.sigma_data**2) ** 0.5,
                    ndim,
                )
                c_out = append_dims(
                    sigma
                    * self.sigma_data
                    / (sigma**2 + self.sigma_data**2) ** 0.5,
                    ndim,
                )
                c_skip = append_dims(
                    self.sigma_data**2 / (sigma**2 + self.sigma_data**2),
                    ndim,
                )
                v = self._uncond_denoise(img * c_in, ts, cond, step, total_step)
                return img * c_skip - v * c_out
            raise ValueError(
                f"unrecognized parameterization `{self.model.parameterization}` occurred"
            )

        s_in = image.new_ones([image.shape[0]])
        return self.sample_step_core(
            image,
            cond,
            step,
            total_step,
            get_denoised,
            quantize=quantize,
            unconditional_cond=unconditional_cond,
            unconditional_guidance_scale=unconditional_guidance_scale,
            **kwargs,
        )

    # internal

    def _t_to_sigma(self, ts: Tensor) -> Tensor:
        log_base = self.log_sigmas_base
        low_idx, high_idx, w = ts.floor().long(), ts.ceil().long(), ts.frac()
        log_sigma = (1.0 - w) * log_base[low_idx] + w * log_base[high_idx]
        return log_sigma.exp()

    def _sigma_to_t(self, sigmas: Tensor, quantize: bool) -> Tensor:
        quantize = self.quantize if quantize is None else quantize
        log_sigmas = sigmas.log()
        dists = log_sigmas - self.log_sigmas_base[:, None]
        if quantize:
            return dists.abs().argmin(dim=0).view(sigmas.shape)
        low_idx = (
            dists.ge(0.0)
            .cumsum(dim=0)
            .argmax(dim=0)
            .clamp(max=self.log_sigmas_base.shape[0] - 2)
        )
        high_idx = low_idx + 1
        low, high = self.log_sigmas_base[low_idx], self.log_sigmas_base[high_idx]
        w = (low - log_sigmas) / (low - high)
        w = w.clamp(0.0, 1.0)
        t = (1.0 - w) * low_idx + w * high_idx
        return t.view(sigmas.shape)

    def _reset_buffers(
        self,
        total_step: int,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
    ) -> None:
        alphas = self.model.alphas_cumprod
        self.sigmas_base = ((1.0 - alphas) / alphas) ** 0.5
        self.log_sigmas_base = self.sigmas_base.log()
        t_max = len(self.sigmas_base) - 1
        ts = torch.linspace(t_max, 0, total_step, device=self.sigmas_base.device)
        self.sigmas = append_zero(self._t_to_sigma(ts)).to(alphas.dtype)
        self.sigma_data = 1.0
        self.quantize = quantize
        # q sampling
        self._q_sampler = KQSampler(self.model)
        self._q_sampler.reset_buffers(self.sigmas.flip(0))
        # unconditional conditioning
        self._reset_uncond_buffers(unconditional_cond, unconditional_guidance_scale)
        # set flag
        self.initialized = True


def klms_coef(order: int, t: np.ndarray, i: int, j: int) -> float:
    def fn(tau: float) -> float:
        prod = 1.0
        for k in range(order):
            if j == k:
                continue
            prod *= (tau - t[i - k]) / (t[i - j] - t[i - k])
        return prod

    if integrate is None:
        raise ValueError("`scipy` is needed for `KLMSSampler`")
    return integrate.quad(fn, t[i], t[i + 1], epsrel=1.0e-4)[0]


@ISampler.register("klms")
class KLMSSampler(KSamplerMixin):
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        get_denoised: IGetDenoised,
        *,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        **kwargs: Any,
    ) -> Tensor:
        order = 4
        denoised = get_denoised(image, self.sigmas[step])
        d = to_d(image, self.sigmas[step], denoised)
        self.ds.append(d)
        if len(self.ds) > order:
            self.ds.pop(0)
        current_order = min(step + 1, order)
        for i, d in enumerate(reversed(self.ds)):
            coef = klms_coef(current_order, self.sigmas_numpy, step, i)
            image = image + coef * d
        return image

    def _reset_buffers(
        self,
        total_step: int,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
    ) -> None:
        super()._reset_buffers(
            total_step,
            quantize,
            unconditional_cond,
            unconditional_guidance_scale,
        )
        self.ds: List[Tensor] = []
        self.sigmas_numpy = self.sigmas.detach().cpu().numpy()


@ISampler.register("k_euler")
class KEulerSampler(KSamplerMixin):
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        get_denoised: IGetDenoised,
        *,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        **kwargs: Any,
    ) -> Tensor:
        s_churn = 0.0
        s_tmin = 0.0
        s_tmax = float("inf")
        s_noise = 1.0
        sigma = self.sigmas[step]
        if not s_tmin <= sigma.item() <= s_tmax:
            gamma = 0.0
        else:
            gamma = min(s_churn / (total_step - 1), 2**0.5 - 1)
        eps = torch.randn_like(image) * s_noise
        sigma_hat = sigma * (gamma + 1)
        if gamma > 0:
            image = image + eps * (sigma_hat**2 - sigma**2) ** 0.5
        denoised = get_denoised(image, sigma_hat)
        d = to_d(image, sigma_hat, denoised)
        dt = self.sigmas[step + 1] - sigma_hat
        image = image + d * dt
        return image


@ISampler.register("k_euler_a")
class KEulerAncestralSampler(KSamplerMixin):
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        get_denoised: IGetDenoised,
        *,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        **kwargs: Any,
    ) -> Tensor:
        eta = 1.0
        s_noise = 1.0
        sigma = self.sigmas[step]
        next_sigma = self.sigmas[step + 1]
        denoised = get_denoised(image, sigma)
        sigma_down, sigma_up = get_ancestral_step(
            sigma.item(),
            next_sigma.item(),
            eta=eta,
        )
        d = to_d(image, sigma, denoised)
        dt = sigma_down - sigma
        image = image + d * dt
        if next_sigma.item() > 0:
            image = image + torch.randn_like(image) * s_noise * sigma_up
        return image


@ISampler.register("k_heun")
class KHeunSampler(KSamplerMixin):
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        get_denoised: IGetDenoised,
        *,
        quantize: bool,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        **kwargs: Any,
    ) -> Tensor:
        s_churn = 0.0
        s_tmin = 0.0
        s_tmax = float("inf")
        s_noise = 1.0
        sigma = self.sigmas[step]
        next_sigma = self.sigmas[step + 1]
        if not s_tmin <= sigma.item() <= s_tmax:
            gamma = 0.0
        else:
            gamma = min(s_churn / (total_step - 1), 2**0.5 - 1)
        eps = torch.randn_like(image) * s_noise
        sigma_hat = sigma * (gamma + 1)
        if gamma > 0:
            image = image + eps * (sigma_hat**2 - sigma**2) ** 0.5
        denoised = get_denoised(image, sigma_hat)
        d = to_d(image, sigma_hat, denoised)
        dt = next_sigma - sigma_hat
        if next_sigma.item() == 0:
            image = image + d * dt
        else:
            x_2 = image + d * dt
            denoised_2 = get_denoised(x_2, next_sigma)
            d_2 = to_d(x_2, next_sigma, denoised_2)
            d_prime = (d + d_2) / 2
            image = image + d_prime * dt
        return image


__all__ = [
    "KLMSSampler",
    "KEulerSampler",
    "KEulerAncestralSampler",
    "KHeunSampler",
]



import torch

import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import List
from typing import Optional
from torch.nn import Module

from ..customs import Linear


class BertPooler(Module):
    def __init__(self, dim: int):
        super().__init__()
        self.linear = nn.Linear(dim, dim)
        self.activation = nn.Tanh()

    def forward(self, net: Tensor) -> Tensor:
        net = net[:, 0]
        net = self.linear(net)
        net = self.activation(net)
        return net


class SequencePooler(Module):
    def __init__(self, dim: int, aux_heads: Optional[List[str]], bias: bool = True):
        super().__init__()
        self.out_dim = 1 + (0 if aux_heads is None else len(aux_heads))
        self.projection = Linear(dim, self.out_dim, bias=bias)

    def forward(self, net: Tensor) -> Tensor:
        weights = self.projection(net)
        weights = F.softmax(weights, dim=1).transpose(-1, -2)
        net = torch.matmul(weights, net)
        if self.out_dim > 1:
            return net
        return net.squeeze(-2)


__all__ = [
    "BertPooler",
    "SequencePooler",
]



from cftool.types import np_dict_type

from ...schema import IRuntimeDataBlock
from ...constants import INPUT_KEY


@IRuntimeDataBlock.register("flatten")
class FlattenBlock(IRuntimeDataBlock):
    def postprocess_item(self, item: np_dict_type) -> np_dict_type:
        item[INPUT_KEY] = item[INPUT_KEY].ravel()
        return item


__all__ = ["FlattenBlock"]



from torch import nn
from torch import Tensor
from typing import Any
from typing import List
from typing import Tuple
from typing import Optional
from torch.nn import Module


TQKV = Tuple[Tensor, Tensor, Tensor]


class IHook(Module):
    # modify the `input`
    ## when `index` is provided, it means the hook is inside a `MultiHooks`
    def before_forward(self, inp: Any, index: Optional[int] = None) -> Any:
        return inp

    # modify the `output`
    def after_forward(self, inp: Any, out: Any) -> Any:
        return out


class IBasicHook(IHook):
    def after_forward(self, inp: Tensor, out: Tensor) -> Tensor:
        return out


class IAttentionHook(IHook):
    def after_forward(self, qkv_inp: TQKV, qkv_out: TQKV) -> TQKV:
        return qkv_out


class MultiHooks(IHook):
    def __init__(self, hooks: List[IHook]) -> None:
        super().__init__()
        self.hooks = nn.ModuleList(hooks)

    def before_forward(self, inp: Any, index: Optional[int] = None) -> Any:
        for i, hook in enumerate(self.hooks):
            inp = hook.before_forward(inp, i)
        return inp

    def after_forward(self, inp: Any, out: Any) -> Any:
        for hook in self.hooks:
            out = hook.after_forward(inp, out)
        return out



import torch

import torch.nn as nn
import torch.nn.functional as F

from typing import Optional
from cftool.types import tensor_dict_type

from ..types import losses_type
from ..schema import ILoss
from ..schema import TrainerState
from ..constants import LOSS_KEY
from ..constants import INPUT_KEY
from ..constants import PREDICTIONS_KEY


MU_KEY = "mu"
LOG_VAR_KEY = "log_var"


@ILoss.register("vae")
@ILoss.register("vae1d")
@ILoss.register("vae2d")
@ILoss.register("siren_vae")
@ILoss.register("style_vae")
class VAELoss(ILoss):
    kld_w: torch.Tensor

    def __init__(
        self,
        *,
        kld_ema: float = 0.999,
        kld_weight: float = 1.0e-3,
    ):
        super().__init__()
        self.kld_ema = kld_ema
        self.kld_weight = kld_weight
        self.register_buffer("kld_w", torch.tensor([0.0], dtype=torch.float32))

    def forward(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
    ) -> losses_type:
        # kld ratio
        if state is not None and not state.is_terminate and self.training:
            eta = self.kld_ema
            ratio = (state.step % state.num_step_per_epoch) / state.num_step_per_epoch
            ratio = (1.0 + ratio) * self.kld_weight
            self.kld_w = eta * self.kld_w + (1.0 - eta) * ratio
        # reconstruction loss
        original = batch[INPUT_KEY]
        reconstruction = forward_results[PREDICTIONS_KEY]
        mse = F.mse_loss(reconstruction, original)
        # kld loss
        mu, log_var = map(forward_results.get, [MU_KEY, LOG_VAR_KEY])
        assert mu is not None and log_var is not None
        var = log_var.exp()
        dim = tuple(i for i in range(1, len(mu.shape)))
        kld_losses = -0.5 * torch.sum(1 + log_var - mu**2 - var, dim=dim)
        kld_loss = torch.mean(kld_losses, dim=0)
        # gather
        if self.training:
            kld_loss = self.kld_w * kld_loss
        loss = mse + kld_loss
        return {"mse": mse, "kld": kld_loss, "kld_w": self.kld_w, LOSS_KEY: loss}


@ILoss.register("vq_vae")
class VQVAELoss(ILoss):
    def __init__(
        self,
        *,
        lb_vq: float = 1.0,
        lb_recon: float = 1.0,
        lb_commit: float = 1.0,
        loss_type: str = "l2",
    ):
        super().__init__()
        self.lb_vq = lb_vq
        self.lb_recon = lb_recon
        self.lb_commit = lb_commit
        self.loss_type = loss_type

    def forward(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        *,
        reduction: str = "mean",
        gather: bool = True,
    ) -> losses_type:
        # reconstruction loss
        original = batch[INPUT_KEY]
        reconstruction = forward_results[PREDICTIONS_KEY]
        if self.loss_type == "l2":
            recon = F.mse_loss(reconstruction, original, reduction=reduction)
        elif self.loss_type == "l1":
            recon = F.l1_loss(reconstruction, original, reduction=reduction)
        else:
            raise ValueError(f"unrecognized loss_type '{self.loss_type}' occurred")
        # vq & commit loss
        z_e = forward_results["z_e"]
        z_q_g = forward_results["z_q_g"]
        vq_loss = F.mse_loss(z_q_g, z_e.detach(), reduction=reduction)
        commit_loss = F.mse_loss(z_e, z_q_g.detach(), reduction=reduction)
        codebook_loss = self.lb_vq * vq_loss + self.lb_commit * commit_loss
        losses = {
            self.loss_type: recon,
            "commit": commit_loss,
            "codebook": codebook_loss,
        }
        if not gather:
            return losses
        # gather
        loss = self.lb_recon * recon + codebook_loss
        losses[LOSS_KEY] = loss
        return losses


__all__ = [
    "MU_KEY",
    "LOG_VAR_KEY",
    "VAELoss",
    "VQVAELoss",
]



import math
import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Callable
from typing import Optional
from torch.nn import Module
from torch.nn import ModuleList
from cftool.misc import filter_kw
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict

from .poolers import BertPooler
from .poolers import SequencePooler
from .channel_mixers import ChannelMixerBase
from .channel_mixers import FeedForward
from .token_mixers import TokenMixerBase
from ..norms import NormFactory
from ..utils import zero_module
from ..common import Lambda
from ..customs import Linear
from ..customs import DropPath
from ..hijacks import HijackConv2d
from ..hijacks import HijackLinear
from ..high_level import PreNorm
from ..attentions import CrossAttention
from ....misc.toolkit import new_seed
from ....misc.toolkit import interpolate
from ....misc.toolkit import gradient_checkpoint


class MixingBlock(Module):
    def __init__(
        self,
        layer_idx: int,
        num_layers: int,
        num_tokens: int,
        in_dim: int,
        latent_dim: int,
        *,
        norm_position: str = "pre_norm",
        token_mixing_type: str,
        token_mixing_config: Optional[Dict[str, Any]] = None,
        token_mixing_dropout: Optional[float] = None,
        channel_mixing_type: str = "ff",
        channel_mixing_config: Optional[Dict[str, Any]] = None,
        dropout: float = 0.0,
        drop_path: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        residual_after_norm: bool = False,
    ):
        def _make_norm() -> nn.Module:
            factory = NormFactory(norm_type)
            return factory.make(in_dim, **(norm_kwargs or {}))

        super().__init__()
        if norm_position not in {"pre_norm", "post_norm"}:
            raise ValueError(
                "`norm_position` should be either 'pre_norm' or 'post_norm', "
                f"'{norm_position}' found"
            )
        self.norm_position = norm_position
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        # token mixing
        if token_mixing_config is None:
            token_mixing_config = {}
        token_mixing_config.update(
            {
                "layer_idx": layer_idx,
                "num_layers": num_layers,
                "num_tokens": num_tokens,
                "in_dim": in_dim,
                "latent_dim": latent_dim,
                "dropout": dropout,
            }
        )
        self.token_norm = _make_norm()
        token_mixing_base = TokenMixerBase.get(token_mixing_type)
        self.token_mixing = safe_execute(token_mixing_base, token_mixing_config)
        if token_mixing_dropout is None:
            token_mixing_dropout = dropout
        self.token_mixing_dropout = nn.Dropout(token_mixing_dropout)
        # channel mixing
        if channel_mixing_config is None:
            channel_mixing_config = {}
        channel_mixing_config.update(
            {
                "layer_idx": layer_idx,
                "num_layers": num_layers,
                "in_dim": in_dim,
                "latent_dim": latent_dim,
                "dropout": dropout,
            }
        )
        self.residual_after_norm = residual_after_norm
        self.channel_norm = _make_norm()
        channel_mixing_base = ChannelMixerBase.get(channel_mixing_type)
        self.channel_mixing = safe_execute(channel_mixing_base, channel_mixing_config)

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        if self.norm_position == "pre_norm":
            return self._pre_norm_forward(net, hw, determinate=determinate, **kwargs)
        if self.norm_position == "post_norm":
            return self._post_norm_forward(net, hw, determinate=determinate, **kwargs)
        raise ValueError(f"unrecognized norm_position '{self.norm_position}' occurred")

    def _pre_norm_forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        # token mixing
        token_mixing_kw = dict(hw=hw, determinate=determinate)
        token_mixing_kw.update(kwargs)
        token_mixing_net = self.token_norm(net)
        token_mixing_net = self.token_mixing(token_mixing_net, **token_mixing_kw)
        token_mixing_net = self.token_mixing_dropout(token_mixing_net)
        net = net + self.drop_path(token_mixing_net)
        # channel mixing
        if self.residual_after_norm:
            net = self.channel_norm(net)
        if not self.channel_mixing.need_2d:
            channel_mixing_net = net
        else:
            if hw is None:
                raise ValueError("`hw` should be provided when FFN needs 2d input")
            channel_mixing_net = net.view(-1, *hw, net.shape[-1])
        if not self.residual_after_norm:
            channel_mixing_net = self.channel_norm(channel_mixing_net)
        channel_mixing_net = self.channel_mixing(channel_mixing_net)
        net = net + self.drop_path(channel_mixing_net)
        return net

    def _post_norm_forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        # token mixing
        token_mixing_kw = dict(hw=hw, determinate=determinate)
        token_mixing_kw.update(kwargs)
        token_mixing_net = self.token_mixing(net, **token_mixing_kw)
        token_mixing_net = self.token_mixing_dropout(token_mixing_net)
        net = net + self.drop_path(token_mixing_net)
        net = self.token_norm(net)
        # channel mixing
        if not self.channel_mixing.need_2d:
            channel_mixing_net = net
        else:
            if hw is None:
                raise ValueError("`hw` should be provided when FFN needs 2d input")
            channel_mixing_net = net.view(-1, *hw, net.shape[-1])
        channel_mixing_net = self.channel_mixing(channel_mixing_net)
        net = net + self.drop_path(channel_mixing_net)
        net = self.channel_norm(net)
        return net


class PositionalEncoding(Module):
    def __init__(
        self,
        dim: int,
        num_tokens: int,
        dropout: float = 0.0,
        *,
        num_head_tokens: int,
        is_vision: bool,
        enable: bool = True,
    ):
        super().__init__()
        self.pos_drop = None
        self.pos_encoding = None
        if enable:
            self.pos_drop = nn.Dropout(p=dropout)
            self.pos_encoding = nn.Parameter(torch.zeros(1, num_tokens, dim))
            nn.init.trunc_normal_(self.pos_encoding, std=0.02)
        self.num_head_tokens = num_head_tokens
        self.is_vision = is_vision

    def forward(
        self,
        net: Tensor,
        *,
        hwp: Optional[Tuple[int, int, int]] = None,
    ) -> Tensor:
        if self.pos_encoding is None or self.pos_drop is None:
            return net
        if self.is_vision:
            feature_span = 0
            pos_encoding = self.interpolate_pos_encoding(net, hwp)
        else:
            feature_span = net.shape[1] - self.num_head_tokens
            pos_encoding = self.pos_encoding[:, :feature_span]
        pos_encoding = self.pos_drop(pos_encoding)
        if self.is_vision or self.num_head_tokens <= 0:
            return net + pos_encoding
        net = net.clone()
        net[:, :feature_span] += pos_encoding
        return net

    # this is for vision positional encodings
    def interpolate_pos_encoding(
        self,
        net: Tensor,
        hwp: Optional[Tuple[int, int, int]],
    ) -> Tensor:
        pos_encoding = self.pos_encoding
        assert pos_encoding is not None
        num_current_history = net.shape[1] - self.num_head_tokens
        num_history = pos_encoding.shape[1] - self.num_head_tokens
        if hwp is None:
            w = h = patch_size = None
        else:
            h, w, patch_size = hwp
        if num_current_history == num_history and w == h:
            return pos_encoding
        if w is None or h is None or patch_size is None:
            raise ValueError("`hwp` should be provided for `interpolate_pos_encoding`")
        head_encoding = None
        if self.num_head_tokens > 0:
            head_encoding = pos_encoding[:, : self.num_head_tokens]
            pos_encoding = pos_encoding[:, self.num_head_tokens :]
        dim = net.shape[-1]
        # This assume that the original input is squared image
        sqrt = math.sqrt(num_history)
        wh_ratio = w / h
        pw = math.sqrt(num_current_history * wh_ratio) + 0.1
        ph = math.sqrt(num_current_history / wh_ratio) + 0.1
        pos_encoding = interpolate(
            pos_encoding.reshape(1, int(sqrt), int(sqrt), dim).permute(0, 3, 1, 2),
            factor=(pw / sqrt, ph / sqrt),
            mode="bicubic",
        )
        assert int(pw) == pos_encoding.shape[-2] and int(ph) == pos_encoding.shape[-1]
        pos_encoding = pos_encoding.permute(0, 2, 3, 1).view(1, -1, dim)
        if head_encoding is None:
            return pos_encoding
        return torch.cat([head_encoding, pos_encoding], dim=1)


class MixedStackedEncoder(Module):
    def __init__(
        self,
        in_dim: int,
        num_tokens: int,
        *,
        token_mixing_type: str,
        token_mixing_config: Optional[Dict[str, Any]] = None,
        channel_mixing_type: str = "ff",
        channel_mixing_config: Optional[Dict[str, Any]] = None,
        num_layers: int = 4,
        dropout: float = 0.0,
        dpr_list: Optional[List[float]] = None,
        drop_path_rate: float = 0.1,
        norm_position: str = "pre_norm",
        norm_type: Optional[str] = "batch_norm",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        embedding_norm: Optional[nn.Module] = None,
        embedding_dropout: Optional[float] = None,
        residual_after_norm: bool = False,
        latent_dim: Optional[int] = None,
        latent_dim_ratio: float = 1.0,
        use_head_token: bool = False,
        head_pooler: Optional[str] = "mean",
        use_positional_encoding: bool = False,
        is_vision_positional_encoding: Optional[bool] = None,
        positional_encoding_dropout: float = 0.0,
        no_head_norm: Optional[bool] = None,
        norm_after_head: bool = False,
        aux_heads: Optional[List[str]] = None,
    ):
        super().__init__()
        # head token
        self.aux_heads = aux_heads
        if not use_head_token:
            num_head_tokens = 0
            self.head_token = None
        else:
            num_head_tokens = 1
            if aux_heads is not None:
                num_head_tokens += len(aux_heads)
            self.head_token = nn.Parameter(torch.zeros(1, num_head_tokens, in_dim))
        self.num_heads = num_head_tokens
        # positional encoding
        num_tokens += num_head_tokens
        if is_vision_positional_encoding is None:
            if use_positional_encoding:
                raise ValueError(
                    "`is_vision_positional_encoding` should be specified when "
                    "`use_positional_encoding` is set to True"
                )
            is_vision_positional_encoding = False
        self.pos_encoding = PositionalEncoding(
            in_dim,
            num_tokens,
            positional_encoding_dropout,
            num_head_tokens=num_head_tokens,
            is_vision=is_vision_positional_encoding,
            enable=use_positional_encoding,
        )
        self.embedding_norm = embedding_norm
        if embedding_dropout is None:
            self.embedding_dropout = None
        else:
            self.embedding_dropout = nn.Dropout(embedding_dropout)
        # core
        if dpr_list is None:
            dpr_list = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]
        if latent_dim is None:
            latent_dim = int(round(in_dim * latent_dim_ratio))
        self.mixing_blocks = ModuleList(
            [
                MixingBlock(
                    i,
                    num_layers,
                    num_tokens,
                    in_dim,
                    latent_dim,
                    norm_position=norm_position,
                    token_mixing_type=token_mixing_type,
                    token_mixing_config=token_mixing_config,
                    channel_mixing_type=channel_mixing_type,
                    channel_mixing_config=channel_mixing_config,
                    dropout=dropout,
                    drop_path=drop_path,
                    norm_type=norm_type,
                    norm_kwargs=norm_kwargs,
                    residual_after_norm=residual_after_norm,
                )
                for i, drop_path in enumerate(dpr_list)
            ]
        )
        # head
        if self.head_token is not None:
            if self.aux_heads is None:
                head = Lambda(lambda x: x[:, 0], name="head_token")
            else:
                head = Lambda(lambda x: x[:, : self.num_heads], name="head_token")
        else:
            if head_pooler is None:
                head = nn.Identity()
            elif head_pooler == "sequence":
                head = SequencePooler(in_dim, aux_heads)
            else:
                if aux_heads is not None:
                    raise ValueError(
                        "either `head_token` or `sequence` head_pooler should be used "
                        f"when `aux_heads` ({aux_heads}) is provided"
                    )
                if head_pooler == "bert":
                    head = BertPooler(in_dim)
                elif head_pooler == "mean":
                    head = Lambda(lambda x: x.mean(1), name="global_average")
                else:
                    msg = f"unrecognized head_pooler '{head_pooler}' occurred"
                    raise ValueError(msg)
        # head norm
        if no_head_norm is None:
            no_head_norm = norm_position == "post_norm"
        if no_head_norm:
            self.head_norm = None
            self.head = head
        elif norm_after_head:
            self.head_norm = NormFactory(norm_type).make(in_dim, **(norm_kwargs or {}))
            self.head = head
        else:
            self.head_norm = None
            self.head = PreNorm(
                in_dim,
                module=head,
                norm_type=norm_type,
                norm_kwargs=norm_kwargs,
            )
        # initializations
        if self.head_token is not None:
            nn.init.trunc_normal_(self.head_token, std=0.02)
        self.apply(self._init_weights)

    def _init_weights(self, m: Module) -> None:
        if isinstance(m, nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0.0)
        elif isinstance(m, Linear):
            m.init_weights_with(lambda t: nn.init.trunc_normal_(t, std=0.02))
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0.0)
            nn.init.constant_(m.weight, 1.0)

    def pre_process(
        self,
        net: Tensor,
        *,
        determinate: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        n, t, d = net.shape
        if self.head_token is not None:
            head_tokens = self.head_token.repeat([n, 1, 1])
            net = torch.cat([head_tokens, net], dim=1)
            t += 1
        if determinate:
            net = net.view(-1, *map(int, [t, d]))
        kw = filter_kw(self.pos_encoding.forward, kwargs)
        net = self.pos_encoding(net, **kw)
        if self.embedding_norm is not None:
            net = self.embedding_norm(net)
        if self.embedding_dropout is not None:
            net = self.embedding_dropout(net)
        return net

    def post_process(self, net: Tensor) -> Tensor:
        net = self.head(net)
        if self.head_norm is not None:
            net = self.head_norm(net)
        return net

    def forward(
        self,
        net: Tensor,
        hw: Optional[Tuple[int, int]] = None,
        **kwargs: Any,
    ) -> Tensor:
        determinate = kwargs.pop("determinate", False)
        net = self.pre_process(net, determinate=determinate, **kwargs)
        for block in self.mixing_blocks:
            net = block(net, hw, determinate=determinate)
        net = self.post_process(net)
        return net


def do_nothing(x: Tensor) -> Tensor:
    return x


def gather(net: Tensor, dim: int, index: Tensor) -> Tensor:
    if net.device.type != "mps" or net.shape[-1] != 1:
        return torch.gather(net, dim, index)
    if dim < 0:
        dim = dim - 1
    return torch.gather(net.unsqueeze(-1), dim, index.unsqueeze(-1)).squeeze(-1)


def bipartite_soft_matching_random2d(
    metric: Tensor,
    w: int,
    h: int,
    sx: int,
    sy: int,
    r: int,
    seed: int,
    no_rand: bool = False,
) -> Tuple[Callable[[Tensor], Tensor], Callable[[Tensor], Tensor]]:
    B, N, _ = metric.shape

    if r <= 0:
        return do_nothing, do_nothing

    with torch.no_grad():
        device = metric.device
        hsy, wsx = h // sy, w // sx

        # For each sy by sx kernel, randomly assign one token to be dst and the rest src
        if no_rand:
            rand_idx = torch.zeros(hsy, wsx, 1, device=device, dtype=torch.int64)
        else:
            generator = torch.Generator(device).manual_seed(seed)
            rand_idx = torch.randint(
                sy * sx, size=(hsy, wsx, 1), generator=generator, device=device
            )

        # The image might not divide sx and sy, so we need to work on a view of the top left if the idx buffer instead
        idx_buffer_view = torch.zeros(
            hsy, wsx, sy * sx, device=device, dtype=torch.int64
        )
        idx_buffer_view.scatter_(dim=2, index=rand_idx, src=-torch.ones_like(rand_idx))
        idx_buffer_view = (
            idx_buffer_view.view(hsy, wsx, sy, sx)
            .transpose(1, 2)
            .reshape(hsy * sy, wsx * sx)
        )

        # Image is not divisible by sx or sy so we need to move it into a new buffer
        if (hsy * sy) < h or (wsx * sx) < w:
            idx_buffer = torch.zeros(h, w, device=device, dtype=torch.int64)
            idx_buffer[: (hsy * sy), : (wsx * sx)] = idx_buffer_view
        else:
            idx_buffer = idx_buffer_view

        # We set dst tokens to be -1 and src to be 0, so an argsort gives us dst|src indices
        rand_idx = idx_buffer.reshape(1, -1, 1).argsort(dim=1)

        # We're finished with these
        del idx_buffer, idx_buffer_view

        # rand_idx is currently dst|src, so split them
        num_dst = hsy * wsx
        a_idx = rand_idx[:, num_dst:, :]  # src
        b_idx = rand_idx[:, :num_dst, :]  # dst

        def split(x: Tensor) -> Tuple[Tensor, Tensor]:
            C = x.shape[-1]
            src = gather(x, dim=1, index=a_idx.expand(B, N - num_dst, C))
            dst = gather(x, dim=1, index=b_idx.expand(B, num_dst, C))
            return src, dst

        # Cosine similarity between A and B
        metric = metric / metric.norm(dim=-1, keepdim=True)
        a, b = split(metric)
        scores = a @ b.transpose(-1, -2)

        # Can't reduce more than the # tokens in src
        r = min(a.shape[1], r)

        # Find the most similar greedily
        node_max, node_idx = scores.max(dim=-1)
        edge_idx = node_max.argsort(dim=-1, descending=True)[..., None]

        unm_idx = edge_idx[..., r:, :]  # Unmerged Tokens
        src_idx = edge_idx[..., :r, :]  # Merged Tokens
        dst_idx = gather(node_idx[..., None], dim=-2, index=src_idx)

    def merge(x: Tensor) -> Tensor:
        src, dst = split(x)
        n, t1, c = src.shape

        unm = gather(src, dim=-2, index=unm_idx.expand(n, t1 - r, c))
        src = gather(src, dim=-2, index=src_idx.expand(n, r, c))
        dst = dst.scatter_reduce(-2, dst_idx.expand(n, r, c), src, reduce="mean")

        return torch.cat([unm, dst], dim=1)

    def unmerge(x: Tensor) -> Tensor:
        unm_len = unm_idx.shape[1]
        unm, dst = x[..., :unm_len, :], x[..., unm_len:, :]
        _, _, c = unm.shape

        src = gather(dst, dim=-2, index=dst_idx.expand(B, r, c))

        # Combine back to the original shape
        out = torch.zeros(B, N, c, device=x.device, dtype=x.dtype)
        out.scatter_(dim=-2, index=b_idx.expand(B, num_dst, c), src=dst)
        out.scatter_(
            dim=-2,
            index=gather(
                a_idx.expand(B, a_idx.shape[1], 1), dim=1, index=unm_idx
            ).expand(B, unm_len, c),
            src=unm,
        )
        out.scatter_(
            dim=-2,
            index=gather(
                a_idx.expand(B, a_idx.shape[1], 1), dim=1, index=src_idx
            ).expand(B, r, c),
            src=src,
        )

        return out

    return merge, unmerge


def compute_merge(x: Tensor, tome_info: Dict[str, Any]) -> Tuple[Callable, ...]:
    original_h, original_w = tome_info["size"]
    original_tokens = original_h * original_w
    downsample = int(math.ceil(math.sqrt(original_tokens // x.shape[1])))

    if downsample > tome_info["max_downsample"]:
        m, u = do_nothing, do_nothing
    else:
        w = int(math.ceil(original_w / downsample))
        h = int(math.ceil(original_h / downsample))
        r = int(x.shape[1] * tome_info["ratio"])
        m, u = bipartite_soft_matching_random2d(  # type: ignore
            x,
            w,
            h,
            tome_info["sx"],
            tome_info["sy"],
            r,
            tome_info.get("seed", new_seed()),
            not tome_info["use_rand"],
        )

    m_a, u_a = (m, u) if tome_info["merge_attn"] else (do_nothing, do_nothing)
    m_c, u_c = (m, u) if tome_info["merge_crossattn"] else (do_nothing, do_nothing)
    m_m, u_m = (m, u) if tome_info["merge_mlp"] else (do_nothing, do_nothing)

    return m_a, m_c, m_m, u_a, u_c, u_m


class SpatialTransformerBlock(Module):
    def __init__(
        self,
        query_dim: int,
        num_heads: int,
        head_dim: int,
        *,
        dropout: float = 0.0,
        context_dim: Optional[int] = None,
        feedforward_multiplier: float = 4.0,
        feedforward_activation: str = "geglu",
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
        tome_info: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.attn1 = CrossAttention(
            query_dim=query_dim,
            num_heads=num_heads,
            head_dim=head_dim,
            dropout=dropout,
            attn_split_chunk=attn_split_chunk,
        )
        latent_dim = round(query_dim * feedforward_multiplier)
        self.ff = FeedForward(
            query_dim,
            latent_dim,
            dropout,
            activation=feedforward_activation,
            add_last_dropout=False,
        )
        self.attn2 = CrossAttention(
            query_dim=query_dim,
            context_dim=context_dim,
            num_heads=num_heads,
            head_dim=head_dim,
            dropout=dropout,
            attn_split_chunk=attn_split_chunk,
        )
        self.norm1 = nn.LayerNorm(query_dim)
        self.norm2 = nn.LayerNorm(query_dim)
        self.norm3 = nn.LayerNorm(query_dim)
        self.use_checkpoint = use_checkpoint
        self.set_tome_info(tome_info)

    # tomesd (https://github.com/dbolya/tomesd)
    def set_tome_info(self, tome_info: Optional[Dict[str, Any]]) -> None:
        if tome_info is None:
            self.tome_info = None
        else:
            self.tome_info = shallow_copy_dict(tome_info)
            self.tome_info.setdefault("ratio", 0.5)
            self.tome_info.setdefault("max_downsample", 1)
            self.tome_info.setdefault("sx", 2)
            self.tome_info.setdefault("sy", 2)
            self.tome_info.setdefault("use_rand", True)
            self.tome_info.setdefault("merge_attn", True)
            self.tome_info.setdefault("merge_crossattn", False)
            self.tome_info.setdefault("merge_mlp", False)

    def forward(self, net: Tensor, context: Optional[Tensor] = None) -> Tensor:
        inputs = (net,) if context is None else (net, context)
        return gradient_checkpoint(
            self._forward,
            inputs=inputs,
            params=self.parameters(),
            enabled=self.use_checkpoint,
        )

    def _forward(self, net: Tensor, context: Optional[Tensor] = None) -> Tensor:
        if self.tome_info is None:
            net = self.attn1(self.norm1(net)) + net
            net = self.attn2(self.norm2(net), context=context) + net
            net = self.ff(self.norm3(net)) + net
        else:
            m_a, m_c, m_m, u_a, u_c, u_m = compute_merge(net, self.tome_info)
            net = u_a(self.attn1(m_a(self.norm1(net)))) + net
            net = u_c(self.attn2(m_c(self.norm2(net)), context=context)) + net
            net = u_m(self.ff(m_m(self.norm3(net)))) + net
        return net


class SpatialTransformer(Module):
    def __init__(
        self,
        in_channels: int,
        num_heads: int,
        head_dim: int,
        *,
        num_layers: int = 1,
        dropout: float = 0.0,
        context_dim: Optional[int] = None,
        use_linear: bool = False,
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
        tome_info: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.norm = nn.GroupNorm(32, in_channels, 1.0e-6, affine=True)
        self.use_linear = use_linear
        latent_channels = num_heads * head_dim
        if not use_linear:
            self.to_latent = HijackConv2d(in_channels, latent_channels, 1, 1, 0)
        else:
            self.to_latent = HijackLinear(in_channels, latent_channels)
        self.blocks = nn.ModuleList(
            [
                SpatialTransformerBlock(
                    latent_channels,
                    num_heads,
                    head_dim,
                    dropout=dropout,
                    context_dim=context_dim,
                    use_checkpoint=use_checkpoint,
                    attn_split_chunk=attn_split_chunk,
                    tome_info=tome_info,
                )
                for _ in range(num_layers)
            ]
        )
        self.from_latent = zero_module(
            HijackConv2d(latent_channels, in_channels, 1, 1, 0)
            if not use_linear
            else HijackLinear(in_channels, latent_channels)
        )

    def set_tome_info(self, tome_info: Optional[Dict[str, Any]]) -> None:
        for block in self.blocks:
            block.set_tome_info(tome_info)

    def forward(self, net: Tensor, context: Optional[Tensor]) -> Tensor:
        inp = net
        b, c, h, w = net.shape
        net = self.norm(net)
        if not self.use_linear:
            net = self.to_latent(net)
        net = net.permute(0, 2, 3, 1).reshape(b, h * w, c)
        if self.use_linear:
            net = self.to_latent(net)
        for block in self.blocks:
            net = block(net, context=context)
        if self.use_linear:
            net = self.from_latent(net)
        net = net.permute(0, 2, 1).contiguous()
        net = net.view(b, c, h, w)
        if not self.use_linear:
            net = self.from_latent(net)
        return inp + net


__all__ = [
    "MixedStackedEncoder",
    "SpatialTransformer",
]



import numpy as np

from PIL import Image
from torch import Tensor
from typing import Any
from typing import Dict
from cftool.types import np_dict_type

from ....schema import IRuntimeDataBlock
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY


@IRuntimeDataBlock.register("to_numpy")
class ToNumpyBlock(IRuntimeDataBlock):
    def postprocess_item(self, item: Dict[str, Any]) -> np_dict_type:
        image = item[INPUT_KEY]
        labels = item[LABEL_KEY]
        if isinstance(image, Tensor):
            image = image.numpy()
        elif isinstance(image, Image.Image):
            if image.mode == "P":
                image = image.convert("RGBA")
            image = np.array(image)
            if len(image.shape) == 2:
                image = image[..., None]
        item[INPUT_KEY] = image
        if isinstance(labels, (int, float)):
            item[LABEL_KEY] = np.array([labels])
        return item


__all__ = ["ToNumpyBlock"]



import torch

from torch import nn
from torch import Tensor
from typing import Optional
from cftool.cv import to_rgb
from cftool.array import l2_normalize
from torchvision.transforms import Resize
from torchvision.transforms import Compose
from torchvision.transforms import ToTensor
from torchvision.transforms import Normalize
from torchvision.transforms import CenterCrop
from torchvision.transforms import InterpolationMode

from .schema import IPerceptor
from ...constants import INPUT_KEY
from ...constants import LATENT_KEY
from ...modules.blocks import HijackLinear
from ..cv.encoder.transformer import ViTEncoder
from ..nlp.encoder.transformer import TeTEncoder


@IPerceptor.register("clip")
class CLIP(IPerceptor):
    vit: Optional[ViTEncoder]
    token_embedding: Optional[nn.Embedding]
    token_type_embedding: Optional[nn.Embedding]
    text_transformer: Optional[TeTEncoder]
    text_latent_dropout: Optional[nn.Dropout]
    text_projection: Optional[HijackLinear]

    def __init__(
        self,
        img_size: int = 224,
        latent_dim: int = 512,
        *,
        # vision
        use_vision: bool = True,
        in_channels: int = 3,
        vision_latent_dim: int = 768,
        vision_patch_size: int = 32,
        vision_num_heads: int = 12,
        vision_num_layers: int = 12,
        vision_norm_eps: float = 1.0e-5,
        vision_feedforward_activation: str = "quick_gelu",
        # text
        use_text: bool = True,
        vocab_size: int = 49408,
        context_length: int = 77,
        use_text_triu_attn_mask: bool = True,
        token_type_size: Optional[int] = None,
        text_latent_dim: int = 512,
        text_padding_idx: int = 0,
        use_text_embedding_norm: bool = False,
        text_embedding_dropout: Optional[bool] = None,
        text_dropout: float = 0.0,
        text_num_heads: int = 8,
        text_num_layers: int = 12,
        text_norm_position: str = "pre_norm",
        text_norm_eps: float = 1.0e-5,
        text_feedforward_activation: str = "quick_gelu",
        text_head_pooler: Optional[str] = None,
    ):
        super().__init__(img_size, context_length)
        # vision
        if not use_vision:
            self.vit = None
        else:
            self._init_vision(
                img_size,
                latent_dim,
                in_channels,
                vision_latent_dim,
                vision_patch_size,
                vision_num_heads,
                vision_num_layers,
                vision_norm_eps,
                vision_feedforward_activation,
            )
        # text
        if not use_text:
            self.token_embedding = None
            self.token_type_embedding = None
            self.text_transformer = None
            self.text_latent_dropout = None
            self.text_projection = None
        else:
            self._init_text(
                latent_dim,
                vocab_size,
                context_length,
                use_text_triu_attn_mask,
                token_type_size,
                text_latent_dim,
                text_padding_idx,
                use_text_embedding_norm,
                text_embedding_dropout,
                text_dropout,
                text_num_heads,
                text_num_layers,
                text_norm_position,
                text_norm_eps,
                text_feedforward_activation,
                text_head_pooler,
            )
        # initialize
        self.reset_parameters()

    def _init_vision(
        self,
        img_size: int,
        latent_dim: int,
        in_channels: int,
        vision_latent_dim: int,
        vision_patch_size: int,
        vision_num_heads: int,
        vision_num_layers: int,
        vision_norm_eps: float,
        vision_feedforward_activation: str,
    ) -> None:
        self.vision_latent_dim = vision_latent_dim
        self.vit = ViTEncoder(
            img_size=img_size,
            patch_size=vision_patch_size,
            in_channels=in_channels,
            latent_dim=self.vision_latent_dim,
            to_patches_config={"bias": False},
            num_layers=vision_num_layers,
            norm_kwargs={"eps": vision_norm_eps},
            embedding_norm=nn.LayerNorm(self.vision_latent_dim, vision_norm_eps),
            attention_kwargs={"num_heads": vision_num_heads},
            feedforward_kwargs={"activation": vision_feedforward_activation},
            norm_after_head=True,
            output_dim=latent_dim,
        )

    def _init_text(
        self,
        latent_dim: int,
        vocab_size: int,
        context_length: int,
        use_text_triu_attn_mask: bool,
        token_type_size: Optional[int],
        text_latent_dim: int,
        text_padding_idx: int,
        use_text_embedding_norm: bool,
        text_embedding_dropout: Optional[bool],
        text_dropout: float,
        text_num_heads: int,
        text_num_layers: int,
        text_norm_position: str,
        text_norm_eps: float,
        text_feedforward_activation: str,
        text_head_pooler: Optional[str],
    ) -> None:
        self.text_num_layers = text_num_layers
        self.text_latent_dim = text_latent_dim
        self.token_embedding = nn.Embedding(
            vocab_size,
            text_latent_dim,
            padding_idx=text_padding_idx,
        )
        if token_type_size is None:
            self.token_type_embedding = None
        else:
            self.token_type_embedding = nn.Embedding(token_type_size, text_latent_dim)
        if not use_text_embedding_norm:
            text_embedding_norm = None
        else:
            text_embedding_norm = nn.LayerNorm(text_latent_dim, text_norm_eps)
        self.text_transformer = TeTEncoder(
            text_latent_dim,
            context_length,
            use_triu_attn_mask=use_text_triu_attn_mask,
            num_layers=text_num_layers,
            dropout=text_dropout,
            norm_position=text_norm_position,
            norm_kwargs={"eps": text_norm_eps},
            embedding_norm=text_embedding_norm,
            embedding_dropout=text_embedding_dropout,
            attention_kwargs={"num_heads": text_num_heads},
            feedforward_kwargs={"activation": text_feedforward_activation},
            head_pooler=text_head_pooler,
        )
        self.text_head_pooler = text_head_pooler
        self.text_latent_dropout = nn.Dropout(text_dropout)
        self.text_projection = HijackLinear(text_latent_dim, latent_dim)

    def reset_parameters(self) -> None:
        if self.token_embedding is not None:
            nn.init.normal_(self.token_embedding.weight, std=0.02)
        tld = self.text_latent_dim
        if self.text_transformer is not None:
            text_encoder = self.text_transformer.encoder
            nn.init.normal_(text_encoder.pos_encoding.pos_encoding, std=0.01)
            proj_std = (tld**-0.5) * ((2 * self.text_num_layers) ** -0.5)
            attn_std = tld**-0.5
            fc_std = (2 * tld) ** -0.5
            for block in text_encoder.mixing_blocks:
                attn = block.token_mixing.net
                mlp = block.channel_mixing.net
                nn.init.normal_(attn.in_w, std=attn_std)
                nn.init.normal_(attn.out_linear.weight, std=proj_std)
                nn.init.normal_(mlp[0].weight, std=fc_std)
                nn.init.normal_(mlp[3].weight, std=proj_std)
        if self.text_projection is not None:
            nn.init.normal_(self.text_projection.weight, std=tld**-0.5)
            nn.init.zeros_(self.text_projection.bias)

    def encode_image(self, image: Tensor) -> Tensor:
        if self.vit is None:
            raise ValueError(
                "`vit` is not initialized, "
                "please set `use_vision=True` when initializing `CLIP`"
            )
        net = self.vit(image, determinate=True)[LATENT_KEY]
        return l2_normalize(net)

    def encode_text(
        self,
        indices: Tensor,
        *,
        apply_pooling: bool = True,
        determinate: bool = True,
        clip_skip: int = 0,
    ) -> Tensor:
        fmt = (
            "`{}` is not initialized, "
            "please set `use_text=True` when initializing `CLIP`"
        )
        if self.token_embedding is None:
            raise ValueError(fmt.format("token_embedding"))
        if self.text_transformer is None:
            raise ValueError(fmt.format("text_transformer"))
        if self.text_latent_dropout is None:
            raise ValueError(fmt.format("text_latent_dropout"))
        if self.text_projection is None:
            raise ValueError(fmt.format("text_projection"))
        net = self.token_embedding(indices)
        if self.token_type_embedding is not None:
            token_type = torch.zeros_like(indices)
            net = net + self.token_type_embedding(token_type)
        kw = dict(clip_skip=clip_skip, determinate=determinate)
        net = self.text_transformer(0, {INPUT_KEY: net}, **kw)[LATENT_KEY]
        if not apply_pooling:
            return net
        # 'pool' the latent net if pooler is not provided
        if self.text_head_pooler is None:
            batch_size = net.shape[0]
            if self.training or batch_size > 1:
                net = net[torch.arange(batch_size), indices.argmax(dim=-1)]
            else:
                # ONNX compatibility
                net = net[:, torch.nonzero(indices)[:, 1][-1]]
        net = self.text_latent_dropout(net)
        net = self.text_projection(net)
        return l2_normalize(net)

    def get_transform(self) -> Compose:
        return Compose(
            [
                Resize(self.img_size, interpolation=InterpolationMode.BICUBIC),
                CenterCrop(self.img_size),
                to_rgb,
                ToTensor(),
                Normalize(
                    (0.48145466, 0.4578275, 0.40821073),
                    (0.26862954, 0.26130258, 0.27577711),
                ),
            ]
        )


__all__ = ["CLIP"]
